/*
 * Tree.java
 *
 * Created on November 13, 2003, 8:33 AM
 */

package edu.umiacs.fmm;

/**
 *
 * @author  wpwy
 */
public class FmmTree{
    public static final int MAX_NUM_LEVEL=8;
    final int dimension = 2;
    int numOfLevels;
    Point[] x;
    Point[] y;
    Potential potential;
    Box[][] struct;
    public static final int DEFAULT_NUM_LEVEL = 3;
    int currLevel;
    
    long numOpsIndirect;
    long numOpsDirect;
    
    /** Creates a new instance of Tree */
    public FmmTree() {
        numOfLevels = DEFAULT_NUM_LEVEL;
        currLevel = numOfLevels-1;
        x = null;
        y = null;
        struct = null;
        potential = new Potential();
    }
    public FmmTree(int level, Potential potential){
        numOfLevels = level;
        currLevel = numOfLevels-1;
        struct = new Box[numOfLevels][];
        for (int i=0; i<numOfLevels; i++) {
            struct[i] = new Box[(int)Math.pow(4,i)];
            for (int j=0; j<struct[i].length; j++)
                struct[i][j] = new Box(i, j, potential.p);
        }
        this.potential = potential;
    }
    public FmmTree(int level, Point[] x, Point[] y, Potential potential){
        //building data structure here
        //level must be >=3;
        numOfLevels = level;
        currLevel = numOfLevels-1;
        struct = new Box[numOfLevels][];
        this.x = x;
        this.y = y;
        this.potential = potential;
        initStruct();
    }
    public FmmTree(Point[] x, Point[] y, Potential potential){
        int N = x.length;
        numOfLevels = Math.round((float)(Math.log(N)/Math.log(2)/2));
        if (numOfLevels<DEFAULT_NUM_LEVEL)
            numOfLevels=DEFAULT_NUM_LEVEL;
        struct = new Box[numOfLevels][];
        this.x = x;
        this.y = y;
        this.potential = potential;
        initStruct();
    }
    private void initStruct() {
        for (int i=0; i<numOfLevels; i++) {
            struct[i] = new Box[(int)Math.pow(4,i)];
            for (int j=0; j<struct[i].length; j++)
                struct[i][j] = new Box(i, j, potential.getP());
        }
        for (int i=0; i<x.length; i++)
            struct[numOfLevels-1][x[i].getBoxIndex(numOfLevels-1)].addX(x[i]);
        for (int i=0; i<y.length; i++)
            struct[numOfLevels-1][y[i].getBoxIndex(numOfLevels-1)].addY(y[i]);
    }
    
    public static FmmTree build(int maxClusterThreshhold, Point[] x, Point[] y, Potential potential) {
        FmmTree ans = null;
        for (int i=DEFAULT_NUM_LEVEL; i<MAX_NUM_LEVEL; i++) {
            ans = new FmmTree(i, x, y, potential);
            //System.out.println("trying numOfLevel: "+ans.numOfLevels);
            if (ans.getClusterThreshold()<=maxClusterThreshhold)
                return ans;
        }
        return ans;
    }
    
    private int getIndex(Point[] z, Point p) {
        int ans = -1;
        for (int i=0; i<z.length; i++)
            if (z[i].equals(p))
                return i;
        return ans;
    }
    
    private void upwardPass(double[] u){
        for (int i=0; i<struct[numOfLevels-1].length; i++) {
            Box thisBox = struct[numOfLevels-1][i];
            Point[] xPoints = thisBox.getX();
            if (xPoints!=null && xPoints.length>0) {
                for (int j=0; j<xPoints.length; j++) {
                    Point thisX = xPoints[j];
                    Complex[] B = potential.getSCoeff(thisX.getCoord(), thisBox.getCenter().getCoord());
                    numOpsIndirect++;
                    numOpsIndirect+=potential.getP();
                    double thisU = u[getIndex(x, thisX)];
                    for (int k=0; k<B.length; k++) {
                        B[k]=B[k].multiply(thisU);
                        numOpsIndirect++;
                    }
                    thisBox.addC(B);
                    numOpsIndirect+=potential.getP();
                }
            }
        }
        
        for (int el = numOfLevels-1; el>=2; el--) {
            //System.out.println("Upward pass level "+el);
            for (int k=0; k<struct[el].length; k++) {
                Box thisBox = struct[el][k];
                //System.out.println(thisBox.toString());
                Box parentBox = thisBox.getParent(this);
                numOpsIndirect++;
                Complex from = thisBox.getCenter().getCoord();
                numOpsIndirect++;
                Complex to = parentBox.getCenter().getCoord();
                numOpsIndirect++;
                parentBox.addC(potential.getSS(from, to, thisBox.getC()));
                numOpsIndirect+=Math.pow(potential.getP(),2);
                numOpsIndirect+=potential.getP();
            }
        }
    }
    
    private void downwardPass1(){
        
        for (int el=2; el<numOfLevels; el++) {
            for (int k=0; k<struct[el].length; k++) {
                
                Box thisBox = struct[el][k];
                //System.out.println("Building Dtilde for "+thisBox.toString());
                Box[] e4 = thisBox.getNeighborsE4(this);
                
                for (int j=0; j<e4.length; j++) {
                    Box thisE4Box = e4[j];
                    
                    Complex from = thisE4Box.getCenter().getCoord();
                    numOpsIndirect++;
                    Complex to = thisBox.getCenter().getCoord();
                    numOpsIndirect++;
                    //Complex[] tmp = thisE4Box.getC();
                    //for (int kk=0; kk<tmp.length; kk++)
                    //    System.out.println(tmp[kk].toString());
                    thisBox.addDtilde( potential.getSR(from, to, thisE4Box.getC()));
                    
                    Complex[] tmpc = potential.getSR(from, to, thisE4Box.getC());
                    
                    //System.out.print(thisE4Box.toString()+" contributed ");
                    //for (Complex tc : tmpc) {
                    //    System.out.print(tc+" ");
                    //}
                    //System.out.println();
                    numOpsIndirect+=Math.pow(potential.getP(),2);
                    numOpsIndirect+=potential.getP();
                    //System.out.println(thisBox.toString());
                }
            }
        }
        
    }
    
    private void downwardPass2(){
        for (int i=0; i<struct[2].length; i++){
            struct[2][i].addD(struct[2][i].getDtilde());
            numOpsIndirect+=potential.getP();
        }
        for (int el=2; el<numOfLevels-1; el++) {
            for (int m=0; m<struct[el].length; m++) {
                Box thisBox = struct[el][m];
                Complex from = thisBox.getCenter().getCoord();
                numOpsIndirect++;
                Box[] children = thisBox.getChildren(this);
                for (int k=0; k<children.length; k++) {
                    Complex to = children[k].getCenter().getCoord();
                    numOpsIndirect++;
                    children[k].addD( potential.getRR(from, to, thisBox.getD()));
                    numOpsIndirect+=Math.pow(potential.getP(),2);
                    children[k].addD( children[k].getDtilde());
                    numOpsIndirect+=potential.getP()*2;
                }
            }
        }
    }
    
    public double[] solve(double[] u) {
        upwardPass(u);
        //System.out.println(this.toString());
        downwardPass1();
        downwardPass2();
        
        double[] v = new double[y.length];
        for (int i=0; i<struct[numOfLevels-1].length; i++) {
            Box thisBox = struct[numOfLevels-1][i];
            Point[] yPoints = thisBox.getY();
            if (yPoints!=null && yPoints.length>0) {
                for (int j=0; j<yPoints.length; j++) {
                    Point thisY = yPoints[j];
                    double regPart = 0d;
                    Complex[] d = thisBox.getD();
                    Complex[] rVec = potential.getRVector(thisY.getCoord(), thisBox.getCenter().getCoord());
                    numOpsIndirect+=potential.getP();
                    for (int k=0; k<d.length; k++){
                        regPart += d[k].multiply(rVec[k]).real();
                        numOpsIndirect++;
                    }
                    
                    double sinPart = 0d;
                    Box[] nei = thisBox.getNeighbors(this);
                    Box[] neiandme = new Box[nei.length+1];
                    for (int z=0; z<nei.length; z++)
                        neiandme[z]=nei[z];
                    neiandme[nei.length]=thisBox;
                    
                    for (int m=0; m<neiandme.length; m++) {
                        Point[] thisNeighborsX = neiandme[m].getX();
                        if (thisNeighborsX!=null && thisNeighborsX.length>0) {
                            for (int q=0; q<thisNeighborsX.length; q++) {
                                Point thisX = thisNeighborsX[q];
                                double thisU = u[getIndex(x, thisX)];
                                double inc = potential.direct(thisY.getCoord(), thisX.getCoord()).multiply(thisU).real();
                                sinPart+=inc;
                                numOpsIndirect++;
                            }
                        }
                    }
                    v[getIndex(y, thisY)] = sinPart+regPart;
                }
            }
        }
        //reinitialize struct();
        //initStruct();
        return v;
    }
    
    public double[] solveDirect(double[] u) {
        double[] v = new double[y.length];
        for (int j=0; j<v.length; j++) {
            for (int i=0; i<x.length; i++) {
                v[j]+=u[i]* potential.direct(y[j].getCoord(), x[i].getCoord()).real();
                numOpsDirect++;
            }
        }
        return v;
    }
    
    public int getClusterThreshold() {
        int ans = 0;
        for (int i=0; i<struct[numOfLevels-1].length; i++) {
            int xlength = struct[numOfLevels-1][i].getX().length;
            int ylength = struct[numOfLevels-1][i].getY().length;
            if (xlength>ans)
                ans = xlength;
            if (ylength>ans)
                ans = ylength;
        }
        return ans;
    }
    
    public Box[][] getStruct(){
        return struct;
    }
    public Box getBox(int level, int index) {
        return struct[level][index];
    }
    public int getNumOfLevels(){
        return this.numOfLevels;
    }
    public String toString() {
        String ans = "";
        for (int i=0; i<numOfLevels; i++) {
            ans+="########Level "+i+": "+"\n";
            for (int j=0; j<struct[i].length; j++){
                ans+=struct[i][j].toString()+"\n";
                ans+="X points: ";
                for (int k=0; k<struct[i][j].getX().length; k++) {
                    ans+=struct[i][j].getX()[k].toString()+" ";
                }
                ans+="\n";
                ans+="Y Points: ";
                for (int k=0; k<struct[i][j].getY().length; k++) {
                    ans+=struct[i][j].getY()[k].toString()+" ";
                }
                ans+="\n";
                ans+="C values: ";
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    ans+=struct[i][j].getC()[k].toString()+" ";
                }
                ans+="\n";
                ans+="Dtilde values: ";
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    ans+=struct[i][j].getDtilde()[k].toString()+" ";
                }
                ans+="\n";
                ans+="D values: ";
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    ans+=struct[i][j].getD()[k].toString()+" ";
                }
                ans+="\n";
            }
            ans+="########End Level "+i+"\n\n";
        }
        return ans;
    }
    public void print(){
        for (int i=0; i<numOfLevels; i++) {
            System.out.println("########Level "+i+": ");
            for (int j=0; j<struct[i].length; j++){
                System.out.println(struct[i][j].toString());
                System.out.print("X points: ");
                for (int k=0; k<struct[i][j].getX().length; k++) {
                    System.out.print(struct[i][j].getX()[k].toString()+" ");
                }
                System.out.println();
                System.out.print("Y Points: ");
                for (int k=0; k<struct[i][j].getY().length; k++) {
                    System.out.print(struct[i][j].getY()[k].toString()+" ");
                }
                System.out.println();
                System.out.print("C values: ");
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    System.out.print(struct[i][j].getC()[k].toString()+" ");
                }
                System.out.println();
                System.out.print("Dtilde values: ");
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    System.out.print(struct[i][j].getDtilde()[k].toString()+" ");
                }
                System.out.println();
                System.out.print("D values: ");
                for (int k=0; k<struct[i][j].getC().length; k++) {
                    System.out.print(struct[i][j].getD()[k].toString()+" ");
                }
                System.out.println();
            }
            System.out.println("########End Level "+i+"\n");
        }
    }
    public Point[] getX() {
        return x;
    }
    public Point[] getY(){
        return y;
    }
    public long getNumOpsDirect(){
        return numOpsDirect;
    }
    public long getNumOps(){
        return getNumOpsIndirect();
    }
    public long getNumOpsIndirect(){
        return numOpsIndirect;
    }
    public void setNumOpsDirect(long i){
        this.numOpsDirect=i;
    }
    public void setNumOpsIndirect(long i){
        this.numOpsIndirect=i;
    }
}

