/*
 * TarWriter.java
 *
 * Created on October 5, 2004, 9:24 AM
 */

package edu.umiacs.fmm.gui;
import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.Color;
import edu.umiacs.fmm.*;
/**
 *
 * @author  wpwy
 */
public class AnimationWorker {
    private int status = Constants.ANIM_STATUS_STOPPED;
    private int command = Constants.ANIM_COMMAND_NOCOMMAND;
    private long interval = Constants.ANIM_INTERVAL_PLAY;
    private LayerAnimation anim = null;
    
    
    /**
     * Utility field used by bound properties.
     */
    private java.beans.PropertyChangeSupport propertyChangeSupport =  new java.beans.PropertyChangeSupport(this);
    
    /**
     * Holds value of property statusMessage.
     */
    private String statusMessage;
    
    /** Creates a new instance of TarWriter */
    public AnimationWorker(LayerAnimation anim) {
        this.anim = anim;
    }
    
    public long getInterval() {
        return interval;
    }
    public void setInterval(long interval){
        this.interval= interval;
    }
    public int getCommand(){
        return command;
    }
    public void setCommand(int c){
        this.command=c;
    }
    
    public void go() {
        final SwingWorker worker = new SwingWorker() {
            public Object construct() {
                setStatus(Constants.ANIM_STATUS_STOPPED);
                setStatusMessage("");
                return new ActualTask();
            }
        };
        worker.start();
    }
    
    public void stop() {
        setCommand(Constants.ANIM_COMMAND_STOP);
        //setStatus(Constants.ANIM_STATUS_STOPPED);
        //setStatusMessage("");
    }
    
    private void processCommandandRepaint() throws AnimationException {
        if (command==Constants.ANIM_COMMAND_PAUSE){
            interval = Constants.ANIM_INTERVAL_PLAY;
            while (command==Constants.ANIM_COMMAND_PAUSE){
                sleepInterval();
            }
        }
        if (command==Constants.ANIM_COMMAND_FF){
            interval = Constants.ANIM_INTERVAL_FF4X;
            sleepInterval();
        } else if (command==Constants.ANIM_COMMAND_FF2END){
            interval = 0;
            //sleepInterval();
        } else if (command==Constants.ANIM_COMMAND_STOP || command==Constants.ANIM_COMMAND_RESET){
            throw new AnimationException();
        }
        if (interval!=0)
            anim.repaint();
    }
    
    //sleeps for a period of time indicated by the current value of "interval"
    private void sleepInterval(){
        sleepInterval(interval);
    }
    //sleeps for a period of time indicated by the i
    private void sleepInterval(long i){
        if (i>0) {
            try {
                Thread.sleep(i);
            } catch(Exception e){
            }
        }
    }
    private void nextStep(){
        //anim.repaint();
        anim.resetHighlightBoxes();
        anim.resetLines();
        if (interval==0) { //FF2END called in this last step
            interval = Constants.ANIM_INTERVAL_PLAY;
            command= Constants.ANIM_COMMAND_PAUSE;
        }
        sleepInterval();
        if (interval!=0)
            anim.repaint();
    }
    
    
    private void animateLines(edu.umiacs.fmm.Point[] from, edu.umiacs.fmm.Point[] to, Color c, int el) throws AnimationException{
        anim.resetLines();
        int numOfLevels = anim.getFmmTree().getStruct().length;
        int lineWidth = numOfLevels-el;
        if (from==null||to==null ||from.length==0||to.length==0)
            return;
        for (int k=0; k<=Constants.ANIM_ARROW_STEPS; k++){
            
            anim.resetLines();
            for (int i=0; i<from.length; i++){
                edu.umiacs.fmm.Point fromPoint=from[i];
                for (int j=0; j<to.length; j++) {
                    edu.umiacs.fmm.Point toPoint=to[j];
                    double fromX = fromPoint.getCoord().real();
                    double fromY = fromPoint.getCoord().imaginary();
                    double toX = toPoint.getCoord().real();
                    double toY = toPoint.getCoord().imaginary();
                    //int fromXint = (int)(fromX*anim.getProportion());
                    //int fromYint = (int)(fromY*anim.getProportion());
                    //int toXint = (int)((fromX+ (toX-fromX)*k/Constants.ANIM_ARROW_STEPS)*anim.getProportion());
                    //int toYint = (int)((fromY+ (toY-fromY)*k/Constants.ANIM_ARROW_STEPS)*anim.getProportion());
                    double toXIntermediate = fromX+ (toX-fromX)*k/Constants.ANIM_ARROW_STEPS;
                    double toYIntermediate = fromY+ (toY-fromY)*k/Constants.ANIM_ARROW_STEPS;
                    anim.addLine(new AnimLine(fromX, fromY, toXIntermediate, toYIntermediate, c, lineWidth));
                }
            }
            sleepInterval();
            processCommandandRepaint();
        }
    }
    private void animateLines(edu.umiacs.fmm.Point[] from, edu.umiacs.fmm.Box to, Color c, int el)throws AnimationException{
        edu.umiacs.fmm.Point[] toArr = new edu.umiacs.fmm.Point[1];
        toArr[0]=to.getCenter();
        animateLines(from, toArr,c, el);
    }
    private void animateLines(edu.umiacs.fmm.Box from, edu.umiacs.fmm.Point[] to, Color c, int el)throws AnimationException {
        edu.umiacs.fmm.Point[] fromArr = new edu.umiacs.fmm.Point[1];
        fromArr[0]=from.getCenter();
        animateLines(fromArr, to,c, el);
    }
    private void animateLines(edu.umiacs.fmm.Box from, edu.umiacs.fmm.Box[] to, Color c, int el)throws AnimationException{
        edu.umiacs.fmm.Point[] fromArr = new edu.umiacs.fmm.Point[1];
        fromArr[0]=from.getCenter();
        edu.umiacs.fmm.Point[] toArr= new edu.umiacs.fmm.Point[to.length];
        for (int i=0; i<to.length; i++)
            toArr[i] = to[i].getCenter();
        animateLines(fromArr, toArr, c, el);
    }
    private void animateLines(edu.umiacs.fmm.Box[] from, edu.umiacs.fmm.Box to, Color c, int el)throws AnimationException{
        edu.umiacs.fmm.Point[] toArr = new edu.umiacs.fmm.Point[1];
        toArr[0]=to.getCenter();
        edu.umiacs.fmm.Point[] fromArr= new edu.umiacs.fmm.Point[from.length];
        for (int i=0; i<from.length; i++)
            fromArr[i] = from[i].getCenter();
        animateLines(fromArr, toArr, c, el);
    }
    private void addCircle(edu.umiacs.fmm.Box b, Color c, int el){
        
        int numOfLevels = anim.getFmmTree().getStruct().length;
        int r = numOfLevels-el;
        int rr = 0;
        if (c.equals(Constants.ANIM_COLOR_S)) {
            r+=4;
            rr=r-2;
        } else if (c.equals(Constants.ANIM_COLOR_SS)) {
            r+=4;
            rr=r-2;
        } else if (c.equals(Constants.ANIM_COLOR_SR)) {
            rr=0;
        } else if (c.equals(Constants.ANIM_COLOR_RR)) {
            r+=2;
            rr=r-2;
        }
        Complex center = b.getCenter().getCoord();
        anim.addCircle(new AnimCircle(center.real(),center.imaginary(), r,rr, c));
    }
    private void addCircles(edu.umiacs.fmm.Box[] b, Color c, int el){
        for (int i=0; i<b.length; i++)
            addCircle(b[i], c, el);
    }
    
    private void runAnimationSExpantion()throws AnimationException{
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing multipole expansion.\n");
        FmmTree fmmTree = this.anim.getFmmTree();
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        for (int i=0; i<struct[numOfLevels-1].length; i++) {
            anim.resetHighlightBoxes();
            edu.umiacs.fmm.Box thisBox = struct[numOfLevels-1][i];
            if (thisBox.getXRecursive(fmmTree).size()>0){
                setStatusMessage("          "+thisBox.toString()+"\n");
                anim.addHighlightBox(thisBox);
                processCommandandRepaint();
                sleepInterval();
                
                animateLines(struct[numOfLevels-1][i].getX(), struct[numOfLevels-1][i], Constants.ANIM_COLOR_S, numOfLevels-1);
                addCircle(struct[numOfLevels-1][i], Constants.ANIM_COLOR_SS, numOfLevels-1);
                processCommandandRepaint();
                sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
            }
        }
        nextStep();
    }
    private void runAnimationSSTranslation()throws AnimationException{
        FmmTree fmmTree = this.anim.getFmmTree();
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing multipole-to-multipole translation.\n");
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        
        for (int el = numOfLevels-2; el>1; el--) {
            //System.out.println("Upward pass level "+el);
            //anim.getStatusPanel().setCurrLevel(el);
            setStatusMessage(".....Level "+el+"\n");
            for (int k=0; k<struct[el].length; k++) {
                anim.resetHighlightBoxes();
                edu.umiacs.fmm.Box thisBox = struct[el][k];
                if (thisBox.getXRecursive(fmmTree).size()>0){
                    setStatusMessage("          "+thisBox.toString()+"\n");
                    processCommandandRepaint();
                    sleepInterval();
                    
                    Complex to = thisBox.getCenter().getCoord();
                    edu.umiacs.fmm.Box[] children = thisBox.getChildren(anim.getFmmTree());
                    anim.addHighlightBoxes(children);
                    Vector<edu.umiacs.fmm.Box> nonEmptyChildren = new Vector<edu.umiacs.fmm.Box>();
                    for (edu.umiacs.fmm.Box child:children)
                        if (child.getXRecursive(fmmTree).size()>0)
                            nonEmptyChildren.addElement(child);
                    animateLines(nonEmptyChildren.toArray(new edu.umiacs.fmm.Box[0]), thisBox, Constants.ANIM_COLOR_SS, el);
                    addCircle(thisBox, Constants.ANIM_COLOR_SS, el);
                    processCommandandRepaint();
                    sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                }
            }
            nextStep();
        }
    }
    
    private void runAnimationDownwardRecursive()throws AnimationException{
        processCommandandRepaint();
        //downward pass recursive
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        
        for (int el = 2; el<numOfLevels-1; el++) {
            
            setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
            setStatusMessage("Performing local-to-local translation.\n");
            setStatusMessage(".....Level: "+el+"\n");
            for (int k=0; k<struct[el].length; k++) {
                //anim.getStatusPanel().setCurrLevel(el);
                //anim.getStatusPanel().setStep(StatusPanel.STEP_RR);
                edu.umiacs.fmm.Box thisBox = struct[el][k];
                if (thisBox.getYRecursive(anim.getFmmTree()).size()>0){
                    setStatusMessage("          "+thisBox.toString()+"\n");
                    edu.umiacs.fmm.Box[] children = thisBox.getChildren(anim.getFmmTree());
                    anim.addHighlightBoxes(children);
                    Vector<edu.umiacs.fmm.Box> nonEmptyChildren = new Vector<edu.umiacs.fmm.Box>();
                    for (edu.umiacs.fmm.Box child:children)
                        if (child.getYRecursive(anim.getFmmTree()).size()>0)
                            nonEmptyChildren.addElement(child);
                    
                    animateLines(thisBox, nonEmptyChildren.toArray(new edu.umiacs.fmm.Box[0]), Constants.ANIM_COLOR_RR, el);
                    addCircles(children, Constants.ANIM_COLOR_RR, el+1);
                    processCommandandRepaint();
                    sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                    
                    anim.resetHighlightBoxes();
                    anim.resetLines();
                }
            }
            
            nextStep();
            processCommandandRepaint();
            
            setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
            setStatusMessage("Performing multipole-to-local translation.\n");
            setStatusMessage(".....Level: "+(el+1)+"\n");
            for (int k=0; k<struct[el+1].length; k++) {
                //anim.getStatusPanel().setCurrLevel(el+1);
                //anim.getStatusPanel().setStep(StatusPanel.STEP_SR);
                edu.umiacs.fmm.Box thisBox = struct[el+1][k];
                if (thisBox.getYRecursive(anim.getFmmTree()).size()>0){
                    setStatusMessage("          "+thisBox.toString()+"\n");
                    Complex to = thisBox.getCenter().getCoord();
                    edu.umiacs.fmm.Box[] e4nei = thisBox.getNeighborsE4(anim.getFmmTree());
                    anim.addHighlightBox(thisBox);
                    anim.addHighlightBoxes(e4nei);
                    processCommandandRepaint();
                    sleepInterval();
                    
                    Vector<edu.umiacs.fmm.Box> e4neiNonEmpty = new Vector<edu.umiacs.fmm.Box>();
                    for (edu.umiacs.fmm.Box bb:e4nei)
                        if (bb.getXRecursive(anim.getFmmTree()).size()>0)
                            e4neiNonEmpty.addElement(bb);
                    animateLines(e4neiNonEmpty.toArray(new edu.umiacs.fmm.Box[0]), thisBox, Constants.ANIM_COLOR_SR, el+1);
                    addCircle(thisBox, Constants.ANIM_COLOR_SR, el+1);
                    processCommandandRepaint();
                    sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                    
                    anim.resetHighlightBoxes();
                    anim.resetLines();
                }
            }
            
            nextStep();
        }
        
    }
    private void runAnimationSRTranslation()throws AnimationException{
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing multipole-to-local translation.\n");
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        //anim.getStatusPanel().setStep(StatusPanel.STEP_SR);
        //anim.getStatusPanel().setCurrLevel(2);
        //downward pass initial phase
        setStatusMessage(".....Level "+2+"\n");
        for (int k=0; k<struct[2].length; k++){
            edu.umiacs.fmm.Box thisBox = struct[2][k];
            if (thisBox.getYRecursive(anim.getFmmTree()).size()>0){
                setStatusMessage("          "+thisBox.toString()+"\n");
                Complex to = thisBox.getCenter().getCoord();
                edu.umiacs.fmm.Box[] e4nei = thisBox.getNeighborsE4(anim.getFmmTree());
                anim.addHighlightBox(thisBox);
                anim.addHighlightBoxes(e4nei);
                processCommandandRepaint();
                sleepInterval();
                Vector<edu.umiacs.fmm.Box> e4neiNonEmpty = new Vector<edu.umiacs.fmm.Box>();
                for (edu.umiacs.fmm.Box bb:e4nei)
                    if (bb.getXRecursive(anim.getFmmTree()).size()>0)
                        e4neiNonEmpty.addElement(bb);
                
                animateLines(e4neiNonEmpty.toArray(new edu.umiacs.fmm.Box[0]), thisBox, Constants.ANIM_COLOR_SR, 2);
                addCircle(thisBox, Constants.ANIM_COLOR_SR, 2);
                processCommandandRepaint();
                sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                
                anim.resetHighlightBoxes();
                anim.resetLines();
            }
        }
        nextStep();
    }
    
    public void runAnimationRExpansion()throws AnimationException{
        processCommandandRepaint();
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        for (int i=0; i<struct[numOfLevels-1].length; i++) {
            anim.resetHighlightBoxes();
            edu.umiacs.fmm.Box thisBox = struct[numOfLevels-1][i];
            if (thisBox.getYRecursive(anim.getFmmTree()).size()>0){
                anim.addHighlightBox(thisBox);
                processCommandandRepaint();
                sleepInterval();
                
                edu.umiacs.fmm.Box thisBox1 = struct[numOfLevels-1][i];
                edu.umiacs.fmm.Point[] targets = struct[numOfLevels-1][i].getY();
                animateLines(thisBox1, targets, Constants.ANIM_COLOR_R, numOfLevels-1);sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                processCommandandRepaint();
                anim.resetLines();
            }
        }
        nextStep();
    }
    
    public void runAnimationFinalSummation() throws AnimationException {
        
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing final summation.\n");
        
        edu.umiacs.fmm.Box[][] struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.length;
        
        //for (int i=0; i<struct[numOfLevels-1].length; i++) {
        for (edu.umiacs.fmm.Box thisBox:struct[numOfLevels-1]){
            if (thisBox.getYRecursive(anim.getFmmTree()).size()>0){
                setStatusMessage("          "+thisBox.toString()+"\n");
                anim.resetHighlightBoxes();
                //edu.umiacs.fmm.Box thisBox = struct[numOfLevels-1][i];
                anim.addHighlightBox(thisBox);
                processCommandandRepaint();
                sleepInterval(interval*Constants.ANIM_ARROW_STEPS);
                
                
                edu.umiacs.fmm.Box[] neighbors = thisBox.getNeighbors(anim.getFmmTree());
                if (neighbors!=null && neighbors.length!=0){
                    for (edu.umiacs.fmm.Box nBox:neighbors)
                        anim.addHighlightBox(nBox);
                }
                processCommandandRepaint();
                sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
            }
        }
        
        nextStep();
    }
    public void runAnimation() throws AnimationException{
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        setStatusMessage("FMM algorithm started.\n");
        setStatus(Constants.ANIM_STATUS_RUNNING);
        sleepInterval();
        //System.out.println("Preparing to do S");
        runAnimationSExpantion();
        //System.out.println("Preparing to do SS");
        runAnimationSSTranslation();
        //System.out.println("Preparing to do SR");
        runAnimationSRTranslation();
        //System.out.println("Preparing to do RR");
        runAnimationDownwardRecursive();
        //System.out.println("Preparing to do R");
        runAnimationRExpansion();
        runAnimationFinalSummation();
        
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        setStatusMessage("FMM algorithm completed\n");
        setStatus(Constants.ANIM_STATUS_COMPLETED);
    }
    
    /**
     * Adds a PropertyChangeListener to the listener list.
     * @param l The listener to add.
     */
    public void addPropertyChangeListener(java.beans.PropertyChangeListener l) {
        
        propertyChangeSupport.addPropertyChangeListener(l);
    }
    
    /**
     * Removes a PropertyChangeListener from the listener list.
     * @param l The listener to remove.
     */
    public void removePropertyChangeListener(java.beans.PropertyChangeListener l) {
        
        propertyChangeSupport.removePropertyChangeListener(l);
    }
    
    /**
     * Getter for property statusMessage.
     * @return Value of property statusMessage.
     */
    public String getStatusMessage() {
        
        return this.statusMessage;
    }
    
    /**
     * Setter for property statusMessage.
     * @param statusMessage New value of property statusMessage.
     */
    public void setStatusMessage(String statusMessage) {
        
        String oldStatusMessage = this.statusMessage;
        this.statusMessage = statusMessage;
        //System.out.println("AnimationWorker: setting status message to "+statusMessage);
        propertyChangeSupport.firePropertyChange("statusMessage", oldStatusMessage, statusMessage);
    }
    
    /**
     * Getter for property status.
     * @return Value of property status.
     */
    public int getStatus() {
        
        return this.status;
    }
    
    /**
     * Setter for property status.
     * @param status New value of property status.
     */
    public void setStatus(int status) {
        
        int oldStatus = this.status;
        this.status = status;
        //System.out.println("AnimationWorker: setting status to "+status);
        propertyChangeSupport.firePropertyChange("status", new Integer(oldStatus), new Integer(status));
    }
    
    /**
     * The actual long running task.  This runs in a SwingWorker thread.
     */
    class ActualTask {
        ActualTask() {
            try{
                runAnimation();
            } catch (AnimationException e){
                //System.out.println("AnimationException CAUGHT");
                setStatus(Constants.ANIM_STATUS_STOPPED);
                setCommand(Constants.ANIM_COMMAND_NOCOMMAND);
            }
        }
        
    }
}
